#include <Windows.h>
#include <ktmw32.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ktmw32")

int main(int argc, char** argv) {
  LSTATUS st;

  //
  // Create two test keys under HKCU\Test.
  //

  HKEY hTestKeyA, hTestKeyB;
  st = RegCreateKeyExA(HKEY_CURRENT_USER,
                       "Test\\AAAA",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hTestKeyA,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA #1 failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hTestKeyA);

  st = RegCreateKeyExA(HKEY_CURRENT_USER,
                       "Test\\BBBB",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hTestKeyB,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA #2 failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hTestKeyB);

  //
  // Create a transaction.
  //

  HANDLE hTrans = CreateTransaction(NULL, NULL, 0, 0, 0, 0, NULL);
  if (hTrans == INVALID_HANDLE_VALUE) {
    printf("CreateTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Open the test keys and rename both of them transactionally to the same new
  // name.
  //

  st = RegOpenKeyTransactedA(HKEY_CURRENT_USER,
                             "Test\\AAAA",
                             0,
                             KEY_ALL_ACCESS,
                             &hTestKeyA,
                             hTrans,
                             NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyTransactedA #1 failed with error %d\n", st);
    return 1;
  }

  st = RegOpenKeyTransactedA(HKEY_CURRENT_USER,
                             "Test\\BBBB",
                             0,
                             KEY_ALL_ACCESS,
                             &hTestKeyB,
                             hTrans,
                             NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyTransactedA #2 failed with error %d\n", st);
    return 1;
  }

  st = RegRenameKey(hTestKeyA, NULL, L"CCCC");

  if (st != ERROR_SUCCESS) {
    printf("RegRenameKey #1 failed with error %d\n", st);
    return 1;
  }

  st = RegRenameKey(hTestKeyB, NULL, L"CCCC");

  if (st != ERROR_SUCCESS) {
    printf("RegRenameKey #2 failed with error %d\n", st);
    return 1;
  }

  //
  // Commit the transaction.
  //

  if (!CommitTransaction(hTrans)) {
    printf("CommitTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Try to recursively delete the test key tree, triggering a crash.
  //

  RegDeleteTree(HKEY_CURRENT_USER, L"Test");

  return 0;
}
